"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.deleteBeer = exports.updateBeer = exports.createBeer = exports.userBeers = exports.getBeers = void 0;
const errorMiddleware_1 = require("../middleware/errorMiddleware");
const express_async_handler_1 = __importDefault(require("express-async-handler"));
const beerModel_1 = require("../models/beerModel");
const authMiddleware_1 = require("../middleware/authMiddleware");
// @desc   Get beers
// @route  GET /api/beers
// @access Private
exports.getBeers = (0, express_async_handler_1.default)((req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    const { page = 1, limit = 100 } = req.query;
    try {
        const [results, itemCount] = yield Promise.all([
            beerModel_1.Beer.find({})
                .sort({ createdAt: -1 })
                .limit(Number(limit))
                .skip((Number(page) - 1) * Number(limit))
                .lean()
                .exec(),
            beerModel_1.Beer.count({}),
        ]);
        const pageCount = Math.ceil(itemCount / Number(limit));
        res.status(200).json({
            currentPage: Number(page),
            itemCount,
            pageCount,
            limitPerPage: Number(limit),
            results,
        });
    }
    catch (err) {
        next(err);
    }
}));
// @desc   User beers
// @route  POST /api/userbeers
// @access Private
exports.userBeers = (0, express_async_handler_1.default)((req, res, next) => __awaiter(void 0, void 0, void 0, function* () {
    const { page = 1, limit = 100 } = req.query;
    try {
        if (!req.header('authorization')) {
            throw new errorMiddleware_1.ValidationError('Not authorized, no token.');
        }
        const decodedToken = (0, authMiddleware_1.decodeToken)(req.header('authorization').split(" ")[1]);
        const userIdFromToken = decodedToken.id;
        const [results, itemCount] = yield Promise.all([
            beerModel_1.Beer.find({ user: userIdFromToken })
                .limit(Number(limit))
                .sort({ createdAt: -1 })
                .skip((Number(page) - 1) * Number(limit))
                .lean()
                .exec(),
            beerModel_1.Beer.count({ user: userIdFromToken }),
        ]);
        const pageCount = Math.ceil(itemCount / Number(limit));
        res.status(200).json({
            currentPage: Number(page),
            itemCount,
            pageCount,
            limitPerPage: Number(limit),
            results,
        });
    }
    catch (error) {
        next(error);
    }
}));
// @desc   Create beers
// @route  POST /api/beers
// @access Private
exports.createBeer = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    if (!req.body.name || !req.body.type || !req.body.rating || !req.body.alcohol || !req.body.description) {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Please fill all fields in form.');
    }
    const beer = yield beerModel_1.Beer.create({
        user: req.user.id,
        createdBy: req.user.name,
        name: req.body.name,
        type: req.body.type,
        rating: req.body.rating,
        alcohol: req.body.alcohol,
        description: req.body.description,
        avatar: req.body.avatar
    });
    yield beer.save();
    res.status(200).json(beer);
}));
// @desc   Update beer
// @route  PUT /api/beers/:id
// @access Private
exports.updateBeer = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const beer = yield beerModel_1.Beer.findById(req.params.id);
    if (!beer) {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Beer not found.');
    }
    const decodeUserId = req.user.id;
    if (beer.user.toString() !== decodeUserId.toString()) {
        res.status(401);
        throw new errorMiddleware_1.ValidationError('Only the beer creator can edit the beer entry.');
    }
    else {
        const updatedBeer = yield beerModel_1.Beer.findByIdAndUpdate(req.params.id, req.body, {
            new: true,
        });
        res.status(200).json(updatedBeer);
    }
}));
// @desc   Delete beer
// @route  DELETE /api/beers/:id
// @access Private
exports.deleteBeer = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const beer = yield beerModel_1.Beer.findById(req.params.id);
    if (!beer) {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Beer not found.');
    }
    const decodeUserId = req.user.id;
    if (beer.user.toString() !== decodeUserId.toString()) {
        res.status(401);
        throw new errorMiddleware_1.ValidationError('Only the beer creator can delete the beer entry.');
    }
    else {
        yield beer.remove();
        res.status(200).json({ id: req.params.id });
    }
}));
//# sourceMappingURL=beerController.js.map