"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getMe = exports.logoutUser = exports.loginUser = exports.registerUser = void 0;
const express_async_handler_1 = __importDefault(require("express-async-handler"));
const bcryptjs_1 = __importDefault(require("bcryptjs"));
const errorMiddleware_1 = require("../middleware/errorMiddleware");
const userModel_1 = require("../models/userModel");
const authMiddleware_1 = require("../middleware/authMiddleware");
// @desc   Register new user
// @route  POST /api/users
// @access Public
exports.registerUser = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const { name, email, password } = req.body;
    if (!name || !email || !password) {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Please fill all fields.');
    }
    //Check if user exist
    const userExists = yield userModel_1.User.findOne({ email });
    if (userExists) {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('User already exists.');
    }
    // Hash password
    const salt = yield bcryptjs_1.default.genSalt(10);
    const hashedPassword = yield bcryptjs_1.default.hash(password, salt);
    // Create user and admin
    const user = yield userModel_1.User.create({
        name,
        email,
        password: hashedPassword,
        role: email === process.env.ADMIN_EMAIL ? 'admin' : 'user',
    });
    const setTokenOnUser = yield userModel_1.User.findOne({ email });
    setTokenOnUser.token = (0, authMiddleware_1.generateToken)(setTokenOnUser._id, setTokenOnUser.role, setTokenOnUser.name, setTokenOnUser.email);
    yield setTokenOnUser.save();
    if (user) {
        res.status(201).json({
            _id: user.id,
            name: user.name,
            email: user.email,
            token: setTokenOnUser.token
        });
    }
    else {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Invalid user data.');
    }
}));
// @desc   Authenticate a user
// @route  POST /api/users/login
// @access Public
exports.loginUser = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const { email, password } = req.body;
    // Check for user email
    const user = yield userModel_1.User.findOne({ email });
    if (user && (yield bcryptjs_1.default.compare(password, user.password))) {
        const token = (0, authMiddleware_1.generateToken)(user._id, user.role, user.name, user.email);
        user.token = token;
        yield user.save();
        res.json({
            _id: user.id,
            name: user.name,
            email: user.email,
            role: user.role,
            token,
        });
    }
    else {
        res.status(400);
        throw new errorMiddleware_1.ValidationError('Invalid email or password.');
    }
}));
exports.logoutUser = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    const decodedToken = (0, authMiddleware_1.decodeToken)(req.header('authorization').split(" ")[1]);
    const userIdFromToken = decodedToken.id;
    // Check for user id
    const user = yield userModel_1.User.findOne({ _id: userIdFromToken });
    user.token = '';
    yield user.save();
    res.status(200).json({
        message: 'User succesfully logged out.'
    });
}));
// @desc   Get user data
// @route  GET /api/users/me
// @access Private
exports.getMe = (0, express_async_handler_1.default)((req, res) => __awaiter(void 0, void 0, void 0, function* () {
    // const userBeers = await Beer.find({user: req.user.id})
    res.status(200).json({
        user: req.user,
        // userBeers,
    });
}));
//# sourceMappingURL=userController.js.map